using System;
using System.Collections.Generic;
using UnityEngine;
using Object = UnityEngine.Object;
using MacGruber;
using SimpleJSON;

namespace CheesyFX
{
    public class Orifice : MonoBehaviour
	{
		private readonly int id;

		public new string name;
		// public List<JSONStorableFloat> bulgeMorphs = new List<JSONStorableFloat>();
		public JSONStorableBool enabledJ = new JSONStorableBool("Enabled", true);

		public JSONStorableFloat arousalRate = new JSONStorableFloat("Arousal Rate", 1f, 0f, 5f);
		
		public JSONStorableFloat bulgeScale = new JSONStorableFloat("Bulge Scale", .4f, 0f, 1f);
		public JSONStorableFloat bulgeDepthScale = new JSONStorableFloat("Bulge Depth Scale", 35f, 0f, 70f);
		public JSONStorableFloat bulgeSharpness = new JSONStorableFloat("Bulge Sharpness", 1f, 0f, 5f);
		
		public JSONStorableFloat triggerOffsetUp = new JSONStorableFloat("Trigger Offset Up", 0f, -.05f, .05f);
		public JSONStorableFloat triggerOffsetForward = new JSONStorableFloat("Trigger Offset Forward", 0f, -.05f, .05f);
		
		public JSONStorableFloat depth = new JSONStorableFloat("Depth", 0f, 0f, .4f, true, false);
		public JSONStorableFloat speed = new JSONStorableFloat("Speed", 0f, 0f, 100f, true, false);
		public JSONStorableFloat stretch = new JSONStorableFloat("Stretch", 0f, 0f, 1f, true, false);
		public JSONStorableFloat relaxation;

		private CustomFloatTrigger depthTrigger;
		private CustomFloatTrigger stretchTrigger;
		private CustomFloatTrigger speedTrigger;
		private CustomFloatTrigger relaxationTrigger;

		protected List<object> UIElements = new List<object>();

		public bool isPenetrated{
			get{return _isPenetrated;}
			set
			{
				if (value == _isPenetrated) return;
				_isPenetrated = value;
				if(!value)
				{
					_tip = null;
					depthMeter.transform.parent = null;
					depth.val = 0f;
				}
			}
		}

		public Atom penetratingAtom
		{
			set
			{
				FillMeUp.penetratingAtoms[this] = value;
			}
			get { return FillMeUp.penetratingAtoms[this]; }
		}
		private bool _isPenetrated;
		public float lastDepth;
		public Transform enterPointTF;
		public Rigidbody depthReference;
		Collider _tip;
		public GameObject depthMeter;
		public Renderer depthMeterRenderer;
		public GameObject enterTriggerGO;
		public Vector3 enterTriggerBasePosition;
		public Collider enterTriggerCollider;
		public TriggerHandler triggerHandler;
		public Func<int, float> bulgeFunc;
		
		public delegate void Actions();
		public Actions onEnterActions;
		public Actions onExitActions;
		private float penetratorWidth;

		public virtual void Init(string name)
		{
			this.name = name;
			depthTrigger = gameObject.AddComponent<CustomFloatTrigger>();
			depthTrigger.Init(FillMeUp.singleton, depth, name);
			stretchTrigger = gameObject.AddComponent<CustomFloatTrigger>();
			stretchTrigger.Init(FillMeUp.singleton, stretch, name);
			speedTrigger = gameObject.AddComponent<CustomFloatTrigger>();
			speedTrigger.Init(FillMeUp.singleton, speed, name);
			if (relaxation != null)
			{
				relaxationTrigger = gameObject.AddComponent<CustomFloatTrigger>();
				relaxationTrigger.Init(FillMeUp.singleton, relaxation, name);
			}
			// bulgeScale.name = name + " "+bulgeScale.name;
			// DAZCharacterSelector characterSelector = bodyRegionManager.atom.GetStorableByID("geometry") as DAZCharacterSelector;
			// GenerateDAZMorphsControlUI morphControl = characterSelector.morphsControlUI;
			// foreach(var s in morphControl.GetMorphUids()){
			// 	if(s.Contains("Belly Bulger")) s.Print();
			// }
			
			
			CreateDepthMeter();
			
			bulgeFunc = x =>
			{
				float exponent = x - (bulgeDepthScale.val * depth.val - 4.4f);
				return bulgeScale.val * penetratorWidth * 4f * Mathf.Pow(1f + bulgeSharpness.val, -exponent * exponent);
			};
			// bulgeScale.setCallbackFunction = x => SetBulgeFunction();
			// bulgeDepthScale.setCallbackFunction = x => SetBulgeFunction();
			// bulgeSharpness.setCallbackFunction = x => SetBulgeFunction();
			
			enabledJ.setCallbackFunction += val => enabled = val;
			triggerOffsetUp.setCallbackFunction = x => UpdateTriggerPosition();
			triggerOffsetForward.setCallbackFunction = x => UpdateTriggerPosition();
			
			// onEnterActions = new Actions();
			// onExitActions = PrintName;
		}

		public void OnDisable()
		{
			Reset();
		}

		public void OnDestroy()
		{
			Destroy(triggerHandler);
			Destroy(depthMeter);
		}

		public void OnAtomRename()
		{
			depthTrigger.OnAtomRename();
			stretchTrigger.OnAtomRename();
			speedTrigger.OnAtomRename();
			if (relaxationTrigger != null) relaxationTrigger.OnAtomRename();
		}

		public virtual JSONClass Store(string subScenePrefix, bool storeTriggers = true)
		{
			JSONClass jc = new JSONClass();
			bulgeScale.Store(jc);
			bulgeSharpness.Store(jc);
			bulgeDepthScale.Store(jc);
			arousalRate.Store(jc);
			if(storeTriggers)
			{
				jc[depthTrigger.name] = depthTrigger.Store(subScenePrefix);
				jc[stretchTrigger.name] = stretchTrigger.Store(subScenePrefix);
				jc[speedTrigger.name] = speedTrigger.Store(subScenePrefix);
				if (relaxationTrigger != null) jc[relaxationTrigger.name] = relaxationTrigger.Store(subScenePrefix);
			}
			return jc;
		}

		public virtual void Load(JSONClass jc, string subScenePrefix)
		{
			if(jc.HasKey(name))
			{
				JSONClass tc = jc[name].AsObject;
				bulgeScale.Load(tc);
				bulgeSharpness.Load(tc);
				bulgeDepthScale.Load(tc);
				arousalRate.Load(tc);
				depthTrigger.Load(tc, subScenePrefix);
				stretchTrigger.Load(tc, subScenePrefix);
				speedTrigger.Load(tc, subScenePrefix);
				if (relaxationTrigger != null) relaxationTrigger.Load(tc, subScenePrefix);
			}
		}

		public void CreateDepthMeter()
		{
			if(depthMeter != null) GameObject.Destroy(depthMeter);
			depthMeter = GameObject.CreatePrimitive(PrimitiveType.Sphere);
			depthMeter.transform.localScale = new Vector3(.005f, .005f, .005f);
			Object.Destroy(depthMeter.GetComponent<Collider>());
			depthMeterRenderer = depthMeter.GetComponent<Renderer>();
			depthMeterRenderer.material.shader = FillMeUp.debugShader;
			switch (name)
			{
				case "Anus":
					depthMeterRenderer.material.color = new Color(1f, 0f, 0.85f, .1f);
					break;
				case "Vagina":
					depthMeterRenderer.material.color = new Color(0f, 0.89f, 1f);
					break;
				default:
					depthMeterRenderer.material.color = new Color(1f, 0.5f, 0f, .2f);
					break;
			}
			depthMeterRenderer.enabled = false;
			depthMeter.tag = "depthMeter";
		}

		void UpdateTriggerPosition()
		{
			enterTriggerGO.transform.localPosition = enterTriggerBasePosition + new Vector3(0f, triggerOffsetUp.val, triggerOffsetForward.val);
		}

		public void GetDepthAndSpeed()
		{
			depth.val = Vector3.Distance(depthMeter.transform.position, enterPointTF.position);
			// depth.val += .02f;
			speed.val = Mathf.Abs(depth.val - lastDepth) / Time.fixedUnscaledDeltaTime*40f;
			lastDepth = depth.val;
			
		}

		public virtual void GetStretch()
		{
			stretch.val = penetratorWidth;
		}

		// public void SetBulgeMorphs1(){
		// 	float mean = bulgeDepthScale.val*depth.val-3f;
		// 	for(int i=0; i<bulgeMorphs.Count; i++){
		// 		bulgeMorphs[i].val = bulgeScale.val*(float)Math.Pow(1f+bulgeSharpness.val, -(i-mean)*(i-mean));
		// 	}
		// }

		public virtual float Stimulate(float scale)
		{
			// float stimulus = arousalRate.val * depth.val * (10f*speed.val + 1f) * (10f*stretch.val + 1f);
			float stimulus = 1f + arousalRate.val * depth.val * (speed.val + 1f) * (stretch.val + 1f);
			FillMeUp.arousalManager.UpdateArousal(stimulus * scale);
			return stimulus * scale;
		}

		// public virtual float StimulateDiscrete()
		// {
		// 	float stimulus = 1f + arousalRate.val * depth.val * (speed.val + 1f) * (stretch.val + 1f);
		// }

		// private void Update()
		// {
		// 	depthTrigger.Trigger(depth.val);
		// 	stretchTrigger.Trigger(stretch.val);
		// 	speedTrigger.Trigger(speed.val);
		// }

		public virtual void FixedUpdate(){
			if(isPenetrated)
			{
				GetDepthAndSpeed();
				GetStretch();
				// orificeManager.bodyRegionManager.arousal.val += arousalRate.val * depth.val * speed.val;
				Stimulate(.05f);
				// depthTrigger.floatTrigger.Update();
				// depthTrigger.Trigger(depth.val);
			}
			else
			{
				if (stretch.val > 0f)
				{
					penetratorWidth = 0f;
					speed.val = 0f;
				}
			}
		}

		public virtual void RegisterCollision1(Collider collider)
		{
			_tip = collider;
			depthMeter.transform.position = enterPointTF.transform.position;
			depthMeter.transform.parent = _tip.transform;
			// depthReference.detectCollisions = false;
		}
		
		public virtual void RegisterCollision(Collider collider)
		{
			float depth = 0f;
			_tip = collider;
			
			Vector3 enterPoint = enterPointTF.position;
			Vector3 deepestPoint = enterPoint;
			float scale = collider.attachedRigidbody.GetAtom().scaleChangeReceivers[0].scale;
			float radius = 0f;
			foreach(var c in collider.attachedRigidbody.transform.GetComponentsInChildren<Collider>())
			{
				// if (c.gameObject.CompareTag("depthMeter")) continue;
				// c.name.Print();
				var capCol = c as CapsuleCollider;
				if(!ReferenceEquals(capCol, null))
				{
					float capColRadius = capCol.radius * scale;
					if (capColRadius > radius) radius = capColRadius;
					Vector3 farthestPoint = enterPoint;
					Transform transform = capCol.transform;
					int dirIndex = capCol.direction;
					Vector3 direction = new Vector3();
					switch (dirIndex)
					{
						case 0:
							direction = transform.right;
							break;
						case 1:
							direction = transform.up;
							break;
						case 2:
							direction = transform.forward;
							break;
					}

					Vector3[] sphereCenters = 
					{
						transform.TransformPoint(capCol.center + direction * (capCol.height * .5f - capCol.radius)),
						transform.TransformPoint(capCol.center - direction * (capCol.height * .5f - capCol.radius))
					};
					float dist = 0f;
					Vector3 farthestSphereCenter = new Vector3();
					bool anySphereCenterInside = false;
					foreach (var center in sphereCenters)
					{
						Vector3 delta = center - enterPoint;
						if (Vector3.Dot(delta, enterPointTF.up) < 0f) continue;
						// c.name.Print();
						if(!anySphereCenterInside) anySphereCenterInside = true;
						float centerDistance = delta.sqrMagnitude;
						if (centerDistance > dist)
						{
							dist = centerDistance;
							farthestSphereCenter = center;
						}
					}

					if(anySphereCenterInside){
						farthestPoint = farthestSphereCenter +
						                (farthestSphereCenter - enterPoint).normalized * capCol.radius;
						float thisDepth = Vector3.Distance(farthestPoint, enterPoint);
						if (thisDepth > depth)
						{
							depth = thisDepth;
							deepestPoint = farthestPoint;
						}
					}
				}
			}
			depthMeter.transform.position = deepestPoint;
			depthMeter.transform.parent = _tip.transform;
			isPenetrated = true;
			penetratorWidth = radius*8.75f;
			// radius.Print();
		}

		public void Reset()
		{
			depth.val = stretch.val = penetratorWidth = 0f;
			depthMeter.transform.parent = null;
		}

		public virtual void CreateUI(){
			enabledJ.CreateUI(UIElements);
			// UIManager.SetupInfoOneLine(UIElements, "Info & Triggers", addSpacer:false);
			// UIManager.SetupInfoOneLine(UIElements, "Settings", addSpacer:false, rightside:true);
			// var tabBar = UIManager.CreateTabBar(new [] {"Triggers & Settings", "Bulge Settings"}, SelectTab, script:FillMeUp.singleton);
			// tabBar.SelectLast();
			// UIElements.Add(tabBar);
		}

		public void ClearUI()
		{
			FillMeUp.singleton.RemoveUIElements(UIElements);
		}

		public void SelectTab(int id)
		{
			FillMeUp.singleton.RemoveUIElements(UIElements);
			// lastMenuId = menuItem;
			if (id == 0) CreateTriggersUI();
			else if(id == 1) CreateBulgeSettingsUI();
		}

		public virtual void CreateTriggersUI()
		{
			var button = FillMeUp.singleton.CreateButton("Depth Triggers");
			button.buttonColor = new Color(0.45f, 1f, 0.45f);
			button.button.onClick.AddListener(() =>
			{
				{
					FillMeUp.singleton.ClearUI();
					depthTrigger.OpenPanel(FillMeUp.singleton, () => FillMeUp.singleton.CreateUI());
				}
			});
			UIElements.Add(button);
			depth.CreateUI(UIElements);
			
			button = FillMeUp.singleton.CreateButton("Stretch Triggers");
			button.buttonColor = new Color(0.45f, 1f, 0.45f);
			button.button.onClick.AddListener(() =>
			{
				{
					FillMeUp.singleton.ClearUI();
					stretchTrigger.OpenPanel(FillMeUp.singleton, () => FillMeUp.singleton.CreateUI());
				}
			});
			UIElements.Add(button);
			stretch.CreateUI(UIElements);
			
			button = FillMeUp.singleton.CreateButton("Speed Triggers");
			button.buttonColor = new Color(0.45f, 1f, 0.45f);
			button.button.onClick.AddListener(() =>
			{
				{
					FillMeUp.singleton.ClearUI();
					speedTrigger.OpenPanel(FillMeUp.singleton, () => FillMeUp.singleton.CreateUI());
				}
			});
			UIElements.Add(button);
			speed.CreateUI(UIElements);
			
			if (relaxation != null)
			{
				button = FillMeUp.singleton.CreateButton("Relaxation Triggers");
				button.buttonColor = new Color(0.45f, 1f, 0.45f);
				button.button.onClick.AddListener(() =>
				{
					{
						FillMeUp.singleton.ClearUI();
						relaxationTrigger.OpenPanel(FillMeUp.singleton, () => FillMeUp.singleton.CreateUI());
					}
				});
				UIElements.Add(button);
				relaxation.CreateUI(UIElements);
			}
		}

		public virtual void CreateBulgeSettingsUI()
		{
			// bulgeScale.CreateUI(UIElements);
			// bulgeSharpness.CreateUI(UIElements, true);
			// bulgeDepthScale.CreateUI(UIElements);
		}
	}
}